/***********************************************************************
 *         libwsmake - Core functionality library of wsmake            *
 *        Copyright (C) 1999,2000,2001,2002 Michael Brownlow           *
 *                                                                     *
 * This program is free software; you can redistribute it and/or modify*
 * it under the terms of the GNU General Public License as published by*
 * the Free Software Foundation; either version 2 of the License, or   *
 * (at your option) any later version.                                 *
 *                                                                     *
 * This program is distributed in the hope that it will be useful,     *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 * GNU General Public License for more details.                        *
 *                                                                     *
 * You should have received a copy of the GNU General Public License   *
 * along with this program; if not, write to the Free Software         *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.           *
 *                                                                     *
 * For questions and comments, please email the author at:             *
 * mike@wsmake.org                                                     *
 ***********************************************************************/
#include "wsLoader.h"
#include "wsSite.h"

#include "osa.h"

#include <unistd.h>
#include <errno.h>

#define MAX_ARGS 256
#define BUF_SZ 1024

extern FILE *yyin;
extern int yyparse();
extern int yylex();

using namespace wsmake;

wsLoader *wsmake::wsLoader_comp = NULL;

char *wsmake::ast_names[] = {
  "Site", "ExpSeq", "DataBlock", "Part", "Data",
  "PrintExp", "AssignExp", "CommandExp", "RegexExp", "SubtagExp",
  "PlusExp", "MinusExp", "DivideExp", "TimesExp", "UminusExp", "UnotExp",
  "PreIncExp", "PostIncExp", "PreDecExp", "PostDecExp", "LshiftExp", "RshiftExp"
};

ast_list *wsmake::cons(ast *a, ast_list *l)
{
  ast_list *ret;

  if(a == NULL) {
    return l;
  } else {
    ret = (ast_list *)osaMalloc(sizeof(ast_list));
    if(ret == NULL) {
      osaPrintError(1, "cons: no memory available, returning original list\n");
      return l;
    }

    ret->a = a;
    ret->next = l;

    return ret;
  }
}

ast *wsmake::mk_node(const ast_kind tag, ast_list *l)
{
  ast *ret;

  ret = (ast *)osaMalloc(sizeof(ast));
  if(ret == NULL) {
    osaPrintError(1, "mk_node: no memory available\n");
    return NULL;
  }

  ret->tag = ast::n_ast;
  ret->n.tag = tag;
  ret->n.l = l;
  if((tag == Site)||
     (tag == ExpSeq)||
     (tag == DataBlock)) {
    ret->line = wsLoader_comp->currblock();
  } else if(tag == Part) { /* Special case for labeled data blocks */
    ret->line = l->a->line;
  } else {
    ret->line = wsLoader_comp->currline();
  }
  ret->filename = wsLoader_comp->currfile();

  return ret;
}

ast* wsmake::mk_int(const long l) {
  ast *ret;

  ret = (ast *)osaMalloc(sizeof(ast));
  if(ret == NULL) {
    osaPrintError(1, "mk_int: no memory available\n");
    return NULL;
  }
  
  ret->tag = ast::i_ast;
  ret->i = l;
  ret->line = wsLoader_comp->currline();
  ret->filename = wsLoader_comp->currfile();

  return ret;
}

ast* wsmake::mk_real(const double r) {
  ast *ret;

  ret = (ast *)osaMalloc(sizeof(ast));
  if(ret == NULL) {
    osaPrintError(1, "mk_real: no memory available\n");
    return NULL;
  }
  
  ret->tag = ast::r_ast;
  ret->r = r;
  ret->line = wsLoader_comp->currline();
  ret->filename = wsLoader_comp->currfile();

  return ret;
}

ast* wsmake::mk_var(const char *v) {
  ast* ret;

  ret = (ast *)osaMalloc(sizeof(ast));
  if(ret == NULL) {
    osaPrintError(1, "mk_var: no memory available\n");
    return NULL;
  }

  ret->tag = ast::v_ast;
  ret->v = osaStrDup(v);
  ret->line = wsLoader_comp->currline();
  ret->filename = wsLoader_comp->currfile();

  return ret;
}

ast* wsmake::mk_str(const char *s) {
  ast* ret;

  ret = (ast *)osaMalloc(sizeof(ast));
  if(ret == NULL) {
    osaPrintError(1, "mk_str: no memory available\n");
    return NULL;
  }

  ret->tag = ast::s_ast;
  ret->s = osaStrDup(s);
  ret->line = wsLoader_comp->currline();
  ret->filename = wsLoader_comp->currfile();

  return ret;
}

void wsmake::ast_dump(ast *a, int level)
{
  ast_list *tmp = NULL;

  if(a == NULL) {
    return;
  }
  switch(a->tag) {
    case ast::i_ast:
      osaPrintf(5, "%*s<ast i=\"%ld\" l=\"%ld\" f=\"%s\"/>\n", level, "",
		a->i, a->line, a->filename);
      break;
    case ast::r_ast:
      osaPrintf(5, "%*s<ast r=\"%f\" l=\"%ld\" f=\"%s\"/>\n", level, "",
		a->r, a->line, a->filename);
      break;
    case ast::v_ast:
      osaPrintf(5, "%*s<ast v=\"%s\" l=\"%ld\" f=\"%s\"/>\n", level, "",
		a->v, a->line, a->filename);
      break;
    case ast::s_ast:
      osaPrintf(5, "%*s<ast s=\"%s\" l=\"%ld\" f=\"%s\"/>\n", level, "",
		a->s, a->line, a->filename);
      break;
    case ast::n_ast:
      osaPrintf(5, "%*s<ast n=\"%s\" l=\"%ld\" f=\"%s\">\n", level, "",
		ast_names[a->n.tag], a->line, a->filename);
      tmp = a->n.l;
      while(tmp != NULL) {
	ast_dump(tmp->a, level+1);
	tmp = tmp->next;
      }
      osaPrintf(5, "%*s</ast> <!-- %s -->\n", level, "",
		ast_names[a->n.tag]);
      break;
    default:
      osaPrintError(1,"%*sast_dump: unknown ast! %d (shouldn't happen)\n",
		level, "", a->tag);
      break;
  };
}

void wsmake::ast_free(ast *a)
{
  ast_list *tmp = NULL;

  if(a == NULL) {
    return;
  }
  switch(a->tag) {
    case ast::i_ast:
      break;
    case ast::r_ast:
      break;
    case ast::v_ast:
      osaFree(a->v);
      break;
    case ast::s_ast:
      osaFree(a->s);
      break;
    case ast::n_ast:
      tmp = a->n.l;
      while(tmp != NULL) {
	ast_free(tmp->a);
	tmp = tmp->next;
      }
      break;
    default:
      osaPrintError(1,"ast_free: unknown ast! %d (shouldn't happen)\n",
		a->tag);
      exit(1);
      break;
  };

  osaFree(a);
}

wsLoader::wsLoader(const list<string> *input_list)
  : line(1), tfile(NULL), block_stack(NULL), block_stack_depth(-1),
    file_stack(NULL), file_stack_depth(-1), file_cache(NULL),
    file_cache_length(-1)
{
  /* Copy input list */
  this->input_list = new list<string>(*input_list);
}

wsLoader::~wsLoader()
{
  /* Free input list memory */
  delete(this->input_list);

  /* FIXME: delete file_cache and block_stack */
}

/* Not reentrant!! */
bool wsLoader::load(void)
{
  bool result = false;
  list<string>::const_iterator i;

  if(wsLoader_comp != NULL) {
    result = false;
  } else {
    wsLoader_comp = this;
    if(this->input_list->size() == 0) {
      osaPrintf(3,"Reading from standard input...\n");
      current_file.assign("stdin");
      memorizeFile("stdin");
      result = (yyparse() == 0)?true:false;
    } else {
      for(i=this->input_list->begin(); i!=this->input_list->end(); i++) {
	osaPrintf(3,"Reading `%s'...\n", i->c_str());
      }
    }
    wsLoader_comp = NULL;
  }

  wsSite::dumpSites();

  return result;
}

void wsLoader::newline(void)
{
  this->line++;
}

void wsLoader::newblock(void)
{
  block_stack_depth++;
  block_stack = (long *)osaRealloc(block_stack,
				   (block_stack_depth+1)*sizeof(long));
  block_stack[block_stack_depth] = currline();
}

void wsLoader::endblock(void)
{
  block_stack_depth--;
  if(block_stack_depth > -1) {
    block_stack = (long *)osaRealloc(block_stack,
				     (block_stack_depth+1)*sizeof(long));
  } else {
    block_stack = (long *)osaFree(block_stack);
  }
}

long wsLoader::currline(void)
{
  return this->line;
}

const char *wsLoader::currfile(void)
{
  return this->rememberFile(this->current_file.c_str());
}

long wsLoader::currblock(void)
{
  if(block_stack_depth == -1) {
    return currline();
  } else {
    return this->block_stack[block_stack_depth];
  }
}

int wsLoader::wrap(void)
{
  fsi *i = NULL;

  osaPrintf(9, "==> Wrapped `%s' (lines=%d)\n", current_file.c_str(),
	    line);

  if(file_stack_depth == -1) {
    return 1;
  } else {
    osaFileClose(yyin);
    if(tfile != NULL) {
      osaFileClose(tfile);
    }
    i = pop();
    ::yyin = i->file;
    current_file.assign(i->filename);
    line = i->line;
    tfile = i->tfile;
    free(i);
    return 0;
  }
}

bool wsLoader::push(const char *filename)
{
  FILE *tf = NULL;

  if(osaFileExists(filename)) {
    if((tf = osaFileOpen(filename, "r")) == NULL) {
      return false;
    }
    file_stack_depth++;
    file_stack = (fsi **)osaRealloc(file_stack,
				    (file_stack_depth+1)*sizeof(struct _fsi *));
    file_stack[file_stack_depth] = (fsi *)osaMalloc(sizeof(struct _fsi));
    file_stack[file_stack_depth]->file = yyin;
    file_stack[file_stack_depth]->filename = osaStrDup(current_file.c_str());
    file_stack[file_stack_depth]->line = line;
    file_stack[file_stack_depth]->tfile = NULL;
    ::yyin = tf;
    current_file.assign(filename);
    memorizeFile(filename);
    line = 0;
    osaPrintf(9,"==> pushed `%s' (old line: %d)\n",
	      file_stack[file_stack_depth]->filename,
	      file_stack[file_stack_depth]->line);
    return true;
  } else {
    return false;
  }
}

fsi *wsLoader::pop(void)
{
  fsi *ret = file_stack[file_stack_depth];
  osaPrintf(9, "==> popped `%s' (new line: %d)\n",
	    file_stack[file_stack_depth]->filename,
	    file_stack[file_stack_depth]->line);
  file_stack_depth--;
  if(file_stack_depth > -1) {
    file_stack = (fsi **)osaRealloc(file_stack,
				    (file_stack_depth+1)*sizeof(struct _fsi *));
  } else {
    file_stack = (fsi **)osaFree(file_stack);
  }
  return ret;
}

void wsLoader::printError(int errtype, const char *topic, ...)
{
  static bool scan = false;
  va_list ap;
  long i;

  if(errtype == e_Parse) {
    if(scan == false) {
      if(file_stack_depth>-1) {
	if(file_stack_depth>0) {
	  osaPrintError(1,"In file included from %s:%d,\n",
			file_stack[file_stack_depth]->filename,
			file_stack[file_stack_depth]->line);
	  for(i=file_stack_depth-1;i>0;i--) {
	    osaPrintError(1,"                 from %s:%d,\n",
			  file_stack[i]->filename, file_stack[i]->line);
	  }
	  osaPrintError(1,"                 from %s:%d:\n",
			file_stack[i]->filename, file_stack[i]->line);
	} else {
	  osaPrintError(1,"In file included from %s:%d:\n",
			file_stack[file_stack_depth]->filename,
			file_stack[file_stack_depth]->line);
	}
      }
      osaPrintError(1, "%s:%d:", current_file.c_str(), line);
      va_start(ap, topic);
      osaVPrintError(1, topic, ap);
      va_end(ap);
      osaPrintError(1, "\n");
    } else {
      scan = false;
    }
  } else {
    scan = true;
    if(file_stack_depth>-1) {
      if(file_stack_depth>0) {
	osaPrintError(1,"In file included from %s:%d,\n",
		      file_stack[file_stack_depth]->filename,
		      file_stack[file_stack_depth]->line);
	for(i=file_stack_depth-1;i>0;i--) {
	  osaPrintError(1,"                 from %s:%d,\n",
			file_stack[i]->filename, file_stack[i]->line);
	}
	osaPrintError(1,"                 from %s:%d:\n",
		      file_stack[i]->filename, file_stack[i]->line);
      } else {
	osaPrintError(1,"In file included from %s:%d:\n",
		      file_stack[file_stack_depth]->filename,
		      file_stack[file_stack_depth]->line);
      }
    }
    osaPrintError(1, "%s:%d:scan error:", current_file.c_str(), line);
    va_start(ap, topic);
    osaVPrintError(1, topic, ap);
    va_end(ap);
    osaPrintError(1, "\n");
  }
}

void wsLoader::splitArgs(char **ap, const char *arg_list)
{
  bool inquotes = false;
  bool indblquotes = false;
  bool lastws = false;      /* was last char whitespace */
  bool lastesc = false;     /* was last char escaped */
  bool isquoted = false;
  char *split_buff = NULL;
  char *bp1 = NULL;   /* buf arg pointer */
  char *bp2 = NULL;   /* buf trav pointer */
  int  cnt = 0;

  split_buff = osaStrDup(arg_list);
  bp1 = bp2 = split_buff;

  while(*bp2 != '\0') {
    if(inquotes) {
      while((*bp2 != '\0')&&(*bp2 != '\'')) bp2++;
      if(*bp2 != '\0') { bp2++; }
      inquotes = false;
    } else if(indblquotes) {
      while((*bp2 != '\0')&&((*bp2 != '"')||(lastesc))) {
	if(*bp2 == '\\') {
	  lastesc = true;
	} else {
	  lastesc = false;
	}
	bp2++;
      }
      if(*bp2 != '\0') { bp2++; }
      indblquotes = false;
    } else {
      if((*bp2 == '\'')&&(!lastesc)) {
	inquotes = true;
	lastws = false;
	isquoted = true;
	bp2++;
      } else if((*bp2 == '"')&&(!lastesc)) {
	indblquotes = true;
	lastws = false;
	isquoted = true;
	bp2++;
      } else if(((*bp2 == ' ')||(*bp2 == '\t'))
		&&(!lastws)&&(cnt < MAX_ARGS)) {
	cnt++;
	*bp2++ = '\0';
	if(isquoted) {
	  bp1++;
	  *(bp2-1) = '\0';
	}
	*ap++ = osaStrDup(bp1);
	bp1 = bp2;
	lastws = true;
	isquoted = false;
      } else {
	if(*bp2 == '\\') {
	  lastesc = true;
	} else {
	  lastesc = false;
	}
	bp2++;
	lastws = false;
      }
    }
  }

  if(isquoted) {
    bp1++;
    *(bp2-1) = '\0';
  }
  *ap++ = osaStrDup(bp1);
  *ap = NULL;

  split_buff = (char *)osaFree(split_buff);
}

void wsLoader::memorizeFile(const char *file)
{
  long i;
  bool found = false;

  /* FIXME: use better search algorithm */
  for(i=-1; i<file_cache_length; i++) {
    if(osaStrCmp(file_cache[file_cache_length], file) == 0) {
      found = true;
      break;
    }
  } 
  
  if(!found) {
    file_cache_length++;
    file_cache = (char **)osaRealloc(file_cache,
				    (file_cache_length+1)*sizeof(char *));
    file_cache[file_cache_length] = osaStrDup(file);
  }
}

const char *wsLoader::rememberFile(const char *file)
{
  long i;

  for(i=-1; i<file_cache_length; i++) {
    if(osaStrCmp(file_cache[file_cache_length], file) == 0) {
      return file_cache[file_cache_length];
    }
  }

  return NULL;
}
